/**
  *  \file planetscentral/console/console.cc
  *  \brief PlanetsCentral Console
  *
  *  This provides a console for PlanetsCentral protocols.
  *
  *  General syntax is:
  *  - a command is a list of words
  *  - words can be quoted with double or single quotes; much like in a
  *    shell. Within a double-quoted word, backslash quotes. The escape
  *    sequences "\n", "\t", "\r", "\0" are supported, as are "\xZZ" and
  *    "\uZZZZ" to produce a raw byte or an UTF-8 encoded character,
  *    respectively. There must be precisely 2 or 4 hex digits. Within an
  *    unquoted or double-quoted word, '$a' is a variable with a one-
  *    character name '${abc}' is a variable with a multi-character name.
  *  - '{' starts a nestable, multiline quote which collects everything
  *    up to a matching '}' in one word. The content would be syntactically
  *    valid c2console code, but no expansion happens in it yet.
  *  - a word starting with/preceded by a '<' is replaced by the content
  *    of the so-named file, for use in e.g.
  *        file put target/file.dat <host/file.dat
  *  - commands can be separated by pipes. The left-hand side command is
  *    executed, and its result appended to the next command. For example,
  *        redis keys foo* | redis del
  *    to delete all keys returned by 'keys foo*',
  *        redis get foo | setenv foo
  *    to get a value from redis and place it in an environment variable.
  *  - comments start with "#" (use this when input is redirected).
  *
  *  Commands are placed in contexts. Entering just a context name will
  *  enter that context and evaluate all following commands within that
  *  context. That is,
  *     file
  *     ls u
  *     get u/file.dat
  *  will be the same as
  *     file ls u
  *     file get u/file.dat
  *  Use "up", "..", or "exit" to leave a context.
  *
  *  Builtin commands:
  *     echo <args>              Echo arguments
  *     setenv <var> <value>     Set variable
  *     checkenv <var...>        Check that variables are set
  *     fatal <command>          Terminate shell if command fails
  *     noerror <command>        Ignore if command fails
  *     silent <command>         Do not print command result
  *     ifset <var> <command>    Execute command only if variable is set
  *     if <cond> <then> [elsif <cond> <elsif>...] [else <else>]
  *                              Parameters are commands (usually, {}-quoted).
  *     foreach <var> <cmd> <values>
  *                              Iteration, usually as 'cmd | foreach...'
  *     macro <name> <commands>  Define macro; callable as <name> <arg1>...<arg20>
  *     file_exists <file...>    Nonzero if all files exist
  *     int <value>              Convert string argument to int
  *     str <value>              Convert int argument to string
  *     env                      Returns list of all variables and values
  *
  *  Commands within redis-protocol contexts (redis, host, file, hostfile, mailout):
  *     repeat <n> <dbcommand>   Execute dbcommand repeatedly, return times
  *     reconnect, reset         Reconnect
  *     exec <dbcommand>         Execute dbcommand (if it starts with a reserved word)
  *     <dbcommand>              Execute dbcommand
  *
  *  Commands within router context:
  *     s <n>                    Set session number
  *     s <n> <sessioncommand>   Send command to session
  *     <sessioncommand>         Send command to session previously set. This applies
  *                              to help, stat, get, post, quit, and save-without-args
  *     <othercommand>           Execute command
  *
  *  \todo Instead of strings, pass real objects around
  */

#include <iostream>
#include <stdexcept>
#include <fstream>
#include <algorithm>
#include <map>
#include <cstdlib>
#include <sstream>
#include <unistd.h>
#include <string.h>
#include "cpluslib/smartptr.h"
#include "cpluslib/stringcv.h"
#include "int/array.h"
#include "int/file.h"
#include "io/streams.h"
#include "planetscentral/console/context.h"
#include "planetscentral/console/router.h"
#include "server/config.h"
#include "util/misc.h"
#include "cpluslib/utf8.h"

static ConnectionContext file_context("file");
static ConnectionContext hostfile_context("hostfile");
static ConnectionContext host_context("host");
static ConnectionContext redis_context("redis");
static ConnectionContext mailout_context("mailout");
static ConnectionContext talk_context("talk");
static RouterContext router_context;

enum TerminalMode { PipeMode, DumbMode, ColorMode };

static Context*
getContext(string_t s)
{
    if (s == "file") {
        return &file_context;
    } else if (s == "hostfile") {
        return &hostfile_context;
    } else if (s == "host") {
        return &host_context;
    } else if (s == "redis") {
        return &redis_context;
    } else if (s == "router") {
        return &router_context;
    } else if (s == "mailout") {
        return &mailout_context;
    } else if (s == "talk") {
        return &talk_context;
    } else {
        return 0;
    }
}

static bool
checkConfig1(string_t key, string_t value)
{
    return file_context.checkConfig("FILE", key, value)
        || hostfile_context.checkConfig("HOSTFILE", key, value)
        || host_context.checkConfig("HOST", key, value)
        || redis_context.checkConfig("REDIS", key, value)
        || mailout_context.checkConfig("MAILOUT", key, value)
        || talk_context.checkConfig("TALK", key, value)
        || router_context.checkConfig(key, value);
}

static void
checkConfig(string_t key, string_t value)
{
    checkConfig1(key, value);
}

class RootContext : public Context {
 public:
    void enter();
    void leave();
    IntValue* processCommand(StringList_t& cmd);
    string_t getPrefix();
};

static RootContext root;
static std::vector<Context*> contexts;
static std::map<string_t, string_t> environment;
static std::map<string_t, string_t> macros;
static TerminalMode terminal_mode;

static IntValue* evalSequence(string_t s);
static bool evalSequenceToBool(string_t s);
static bool evalStream(std::istream& in, TerminalMode terminal_mode);

void
RootContext::enter()
{ }

void
RootContext::leave()
{ }

IntValue*
RootContext::processCommand(StringList_t& cmd)
{
    if (Context* ctx = getContext(cmd[0])) {
        if (cmd.size() != 1) {
            cmd.erase(cmd.begin());
            IntValue* result;
            ctx->enter();
            result = ctx->processCommand(cmd);
            ctx->leave();
            return result;
        } else {
            contexts.push_back(ctx);
            ctx->enter();
            return 0;
        }
    } else {
        throw std::runtime_error("Unknown command");
    }
}

string_t
RootContext::getPrefix()
{
    return "c2console";
}

static void
showValue(IntValue* v, string_t prefix)
{
    if (!v) {
        std::cout << "null";
    } else if (IntArray* av = dynamic_cast<IntArray*>(v)) {
        IntDataSegment& seg = av->getData()->content;
        std::cout << "[";
        for (uint32_t i = 0; i < seg.size(); ++i) {
            std::cout << "\n" << prefix << "  ";
            showValue(seg[i], prefix + "  ");
            if (i+1 < seg.size()) {
                std::cout << ",";
            }
        }
        if (seg.size()) {
            std::cout << "\n" << prefix << "]";
        } else {
            std::cout << " ]";
        }
    } else {
        std::cout << v->toString(true);
    }
}

static void
unsetEnvironment(const string_t& key, const string_t oldValue, bool hadOldValue)
{
    if (hadOldValue) {
        environment[key] = oldValue;
    } else {
        environment.erase(key);
    }
}

static IntValue*
evalCommand(StringList_t& cmd)
{
    std::map<string_t, string_t>::iterator it = macros.find(cmd[0]);
    if (it != macros.end()) {
        // Macro invocation. Pass args.
        enum { MAX_ARGS = 20 };
        bool hadOldValue[MAX_ARGS];
        string_t oldValues[MAX_ARGS];
        for (uint_t i = 1; i <= MAX_ARGS; ++i) {
            string_t name = numToString(i);
            std::map<string_t, string_t>::iterator old = environment.find(name);
            if (old != environment.end()) {
                hadOldValue[i-1] = true;
                oldValues[i-1] = old->second;
            } else {
                hadOldValue[i-1] = false;
            }
            if (i < cmd.size()) {
                environment[name] = cmd[i];
            } else {
                environment.erase(name);
            }
        }

        // Call it.
        try {
            std::auto_ptr<IntValue> result(evalSequence(it->second));
            for (uint_t i = 1; i <= MAX_ARGS; ++i) {
                unsetEnvironment(numToString(i), oldValues[i-1], hadOldValue[i-1]);
            }
            return result.release();
        }
        catch (...) {
            for (uint_t i = 1; i <= MAX_ARGS; ++i) {
                unsetEnvironment(numToString(i), oldValues[i-1], hadOldValue[i-1]);
            }
            throw;
        }
    } else if (cmd[0] == ".." || cmd[0] == "up" || cmd[0] == "exit") {
        contexts.back()->leave();
        contexts.pop_back();
        return 0;
    } else if (cmd[0] == "echo") {
        for (StringList_t::size_type i = 1; i < cmd.size(); ++i) {
            if (i != 1) {
                std::cout << " ";
            }
            std::cout << cmd[i];
        }
        std::cout << "\n";
        return 0;
    } else if (cmd[0] == "die") {
        for (StringList_t::size_type i = 1; i < cmd.size(); ++i) {
            if (i != 1) {
                std::cerr << " ";
            }
            std::cerr << cmd[i];
        }
        std::cerr << "\n";
        std::exit(1);
    } else if (cmd[0] == "setenv") {
        if (cmd.size() != 3) {
            throw std::runtime_error("Argument error in 'setenv'");
        }
        environment[cmd[1]] = cmd[2];
        return 0;
    } else if (cmd[0] == "checkenv") {
        for (uint32_t i = 1; i < cmd.size(); ++i) {
            if (environment.find(cmd[i]) == environment.end()) {
                throw std::runtime_error("Required environment variable is missing");
            }
        }
        return 0;
    } else if (cmd[0] == "fatal") {
        cmd.erase(cmd.begin());
        try {
            if (cmd.empty()) {
                throw std::runtime_error("Need an argument for 'fatal'");
            }
            return evalCommand(cmd);
        }
        catch (std::exception& e) {
            std::cerr << "ERROR: " << e.what() << "\n";
            std::exit(1);
        }
    } else if (cmd[0] == "noerror") {
        cmd.erase(cmd.begin());
        try {
            if (!cmd.empty()) {
                return evalCommand(cmd);
            }
        }
        catch (std::exception& e) { }
        return 0;
    } else if (cmd[0] == "silent") {
        cmd.erase(cmd.begin());
        delete evalCommand(cmd);
        return 0;
    } else if (cmd[0] == "ifset") {
        if (cmd.size() < 3) {
            throw std::runtime_error("Too few arguments for 'ifset'");
        }
        std::map<string_t, string_t>::iterator i = environment.find(cmd[1]);
        if (i != environment.end() && !i->second.empty()) {
            cmd.erase(cmd.begin());
            cmd.erase(cmd.begin());
            return evalCommand(cmd);
        } else {
            return 0;
        }
    } else if (cmd[0] == "foreach") {
        // foreach VAR BODY ITEMS...
        if (cmd.size() < 3) {
            throw std::runtime_error("Too few arguments for 'foreach'");
        }
        string_t oldValue;
        bool hadOldValue = false;
        std::map<string_t, string_t>::iterator i = environment.find(cmd[1]);
        if (i != environment.end()) {
            hadOldValue = true;
            oldValue = i->second;
        }
        try {
            for (uint32_t i = 3; i < cmd.size(); ++i) {
                environment[cmd[1]] = cmd[i];
                delete evalSequence(cmd[2]);
            }
            unsetEnvironment(cmd[1], oldValue, hadOldValue);
        }
        catch (...) {
            unsetEnvironment(cmd[1], oldValue, hadOldValue);
            throw;
        }
        return 0;
    } else if (cmd[0] == "if") {
        // if CONDITION THEN [elsif CONDITION ELSIF] [else ELSE]
        uint32_t execIndex = 0;
        uint32_t index = 1;

        // Then part
        if (cmd.size() < 3) {
            throw std::runtime_error("Too few arguments for 'if'");
        }
        if (evalSequenceToBool(cmd[1])) {
            execIndex = 2;
        }
        index = 3;

        // Elsif parts
        while (cmd.size() > index && cmd[index] == "elsif") {
            if (cmd.size() < index+3) {
                throw std::runtime_error("Too few arguments for 'if'");
            }
            if (execIndex == 0 && evalSequenceToBool(cmd[index+1])) {
                execIndex = index+2;
            }
            index += 3;
        }

        // Else part
        if (cmd.size() > index && cmd[index] == "else") {
            if (cmd.size() < index+2) {
                throw std::runtime_error("Too few arguments for 'if'");
            }
            if (execIndex == 0) {
                execIndex = index+1;
            }
            index += 2;
        }

        if (cmd.size() > index) {
            throw std::runtime_error("Too many arguments for 'if'");
        }

        // Do it
        if (execIndex != 0) {
            delete evalSequence(cmd[execIndex]);
        }
        return 0;
    } else if (cmd[0] == "macro") {
        // macro NAME BODY
        if (cmd.size() != 3) {
            throw std::runtime_error("'macro' needs two arguments");
        }
        macros[cmd[1]] = cmd[2];
        return 0;
    } else if (cmd[0] == "file_exists") {
        int result = 1;
        for (uint32_t i = 1; i < cmd.size(); ++i) {
            if (!Stream::openFileNT(cmd[i], Stream::C_READ)) {
                result = 0;
                break;
            }
        }
        return makeIntValue(result);
    } else if (cmd[0] == "int") {
        if (cmd.size() != 2) {
            throw std::runtime_error("Need one argument for 'int'");
        }
        int32_t n;
        if (stringValue(cmd[1], n)) {
            return makeIntValue(n);
        } else {
            return 0;
        }
    } else if (cmd[0] == "int_not") {
        if (cmd.size() != 2) {
            throw std::runtime_error("Need one argument for 'int_not'");
        }
        int32_t n;
        if (stringValue(cmd[1], n)) {
            return makeIntValue(!n);
        } else {
            return 0;
        }
    } else if (cmd[0] == "int_add") {
        int32_t sum = 0;
        for (uint32_t i = 1; i < cmd.size(); ++i) {
            int32_t n;
            if (!stringValue(cmd[i], n)) {
                throw std::runtime_error("Need integer argument to 'int_add'");
            }
            sum += n;
        }
        return makeIntValue(sum);
    } else if (cmd[0] == "int_seq") {
        if (cmd.size() != 3) {
            throw std::runtime_error("Need two arguments for 'int_seq'");
        }
        int32_t low, high;
        if (!stringValue(cmd[1], low) || !stringValue(cmd[2], high)) {
            throw std::runtime_error("Need integer arguments for 'int_seq'");
        }
        Ptr<IntArrayData> ad = new IntArrayData();
        while (low <= high) {
            ad->content.pushNew(makeIntValue(low));
            ++low;
        }
        ad->addDimension(ad->content.size());
        return new IntArray(ad);
    } else if (cmd[0] == "str") {
        if (cmd.size() != 2) {
            throw std::runtime_error("Need one argument for 'str'");
        }
        return makeStringValue(cmd[1]);
    } else if (cmd[0] == "str_empty") {
        bool result = true;
        for (uint_t i = 1; i < cmd.size(); ++i) {
            if (!cmd[i].empty()) {
                result = false;
                break;
            }
        }
        return makeBoolValue(result);
    } else if (cmd[0] == "env") {
        Ptr<IntArrayData> data = new IntArrayData();
        for (std::map<string_t, string_t>::iterator i = environment.begin(); i != environment.end(); ++i) {
            data->content.push(makeStringValue(i->first));
            data->content.push(makeStringValue(i->second));
        }
        data->addDimension(data->content.size());
        return new IntArray(data);
    } else if (cmd[0] == "load") {
        for (uint_t i = 1; i < cmd.size(); ++i) {
            std::ifstream in(cmd[i].c_str());
            if (!in) {
                string_t msg("Unable to open '");
                msg += cmd[i];
                msg += "'";
                throw std::runtime_error(msg.c_str());
            }
            evalStream(in, PipeMode);
        }
        return 0;
    } else {
        return contexts.back()->processCommand(cmd);
    }
}

static void
handleLE(StringList_t& cmd, bool& hadLE)
{
    if (hadLE && cmd.size()) {
        try {
            Ptr<Stream> s = Stream::openFile(cmd.back(), Stream::C_READ);
            char tmp[1024];
            string_t result;
            while (int n = s->read(tmp, sizeof(tmp))) {
                result.append(tmp, n);
            }
            std::cout << "(loaded " << cmd.back() << ", " << result.size() << " bytes)\n";
            cmd.back() = result;
        }
        catch (std::exception& e) {
            if (cmd[0] == "fatal") {
                std::cout << "ERROR: " << e.what() << "\n";
                std::exit(1);
            }
            throw;
        }
    }
    hadLE = false;
}

static void
handlePipe(StringList_t& thisCommand, StringList_t& previousPipeResult)
{
    if (thisCommand.size() == 0) {
        throw std::runtime_error("No verb in command");
    }
    for (StringList_t::iterator i = previousPipeResult.begin(); i != previousPipeResult.end(); ++i) {
        thisCommand.push_back(*i);
    }
    previousPipeResult.clear();
}

static bool
handleHexDigit(int32_t& value, char c)
{
    if (c >= '0' && c <= '9') {
        value = 16*value + c - '0';
        return true;
    } else if (c >= 'A' && c <= 'F') {
        value = 16*value + c - 'A' + 10;
        return true;
    } else if (c >= 'a' && c <= 'f') {
        value = 16*value + c - 'a' + 10;
        return true;
    } else {
        return false;
    }
}

/** Evaluate a command.
    \param in Stream to read from
    \param terminal_mode terminal mode
    \param line Line to evaluate
    \param lastResult [in/out] Last effective result
    \retval true a command was executed
    \retval false this was a blank line */
static bool
evalLine(std::istream& in, TerminalMode terminal_mode, string_t line, std::auto_ptr<IntValue>& lastResult)
{
    // Parse it
    StringList_t previousPipeResult;
    StringList_t thisCommand;
    enum State {
        Blank,                  // in whitespace, next character starts word
        Word,                   // in word
        WordVar,                // '$' in word
        WordVarBrace,           // '${' in word
        DQ,                     // in double-quoted word
        DQQ,                    // in double-quoted word, after backslash
        DQX,                    // in double-quoted word, after '\x'
        DQU,                    // in double-quoted word, after '\u'
        DQVar,                  // '$' in double-quoted
        DQVarBrace,             // '${' in double-quoted
        SQ,                     // in single-quoted word
        LE,                     // after '<'
        Brace,                  // '{' block
        BraceDQ,                // '{"
        BraceDQQ,               // '{"\'
        BraceSQ,                // '{''
        Pipe                    // after pipe (temporary state)
    };
    bool hadLE = false;         // this word had a '<' in front
    bool hadNonblank = false;
    State s = Blank;
    string_t varName;
    int32_t braceLevel = 0;
    int32_t charValue = 0;
    int charDigits = 0;
 again:
    for (string_t::size_type i = 0; i < line.size(); ++i) {
        // Get character
        uint8_t ch = line[i];
        State chClass = (ch == ' ' || ch == '\r' || ch == '\n')
            ? Blank
            : ch == '"'
            ? DQ
            : ch == '\''
            ? SQ
            : ch == '<'
            ? LE
            : ch == '|'
            ? Pipe
            : ch == '$'
            ? WordVar
            : ch == '{'
            ? Brace
            : Word;

        // Basic syntax checks
        if (ch == '#' && (s == Blank || s == Word || s == LE)) {
            break;
        }

        if (chClass != Blank) {
            hadNonblank = true;
        }

        // Process it
        switch (s) {
         case Blank:
            if (chClass == Pipe || chClass == LE || chClass == Blank) {
                // State change without action
            } else if (chClass == Word) {
                // Normal word
                thisCommand.push_back(string_t(1, uint8_t(ch)));
            } else {
                // Quoted word or variable
                thisCommand.push_back(string_t());
            }
            s = chClass;
            break;

         case Word:
            if (chClass == Pipe || chClass == LE || chClass == Blank) {
                // End word
                handleLE(thisCommand, hadLE);
                s = chClass;
            } else if (chClass == Word) {
                // Word char
                thisCommand.back() += char(ch);
            } else {
                // Quote or variable
            }
            s = chClass;
            break;

         case DQ:
            if (ch == '"') {
                s = Word;
            } else if (ch == '\\') {
                s = DQQ;
            } else if (ch == '$') {
                s = DQVar;
            } else {
                thisCommand.back() += char(ch);
            }
            break;

         case DQQ:
            if (ch == 'x') {
                charValue = 0;
                charDigits = 0;
                s = DQX;
            } else if (ch == 'u') {
                charValue = 0;
                charDigits = 0;
                s = DQU;
            } else {
                if (ch == '0') {
                    thisCommand.back() += '\0';
                } else if (ch == 'n') {
                    thisCommand.back() += '\n';
                } else if (ch == 't') {
                    thisCommand.back() += '\t';
                } else if (ch == 'r') {
                    thisCommand.back() += '\r';
                } else {
                    thisCommand.back() += char(ch);
                }
                s = DQ;
            }
            break;

         case DQX:
            if (handleHexDigit(charValue, ch)) {
                ++charDigits;
                if (charDigits >= 2) {
                    thisCommand.back() += char(charValue);
                    s = DQ;
                }
            } else {
                throw std::runtime_error("Expecting hex digit after '\\x'");
            }
            break;

         case DQU:
            if (handleHexDigit(charValue, ch)) {
                ++charDigits;
                if (charDigits >= 4) {
                    appendUtf8(thisCommand.back(), charValue);
                    s = DQ;
                }
            } else {
                throw std::runtime_error("Expecting hex digit after '\\u'");
            }
            break;

         case SQ:
            if (ch == '\'') {
                s = Word;
            } else {
                thisCommand.back() += char(ch);
            }
            break;

         case LE:
            if (chClass == Blank) {
                // ok
            } else if (chClass == Pipe || chClass == LE) {
                throw std::runtime_error("Expecting name after '<'");
            } else {
                if (chClass == Word) {
                    thisCommand.push_back(string_t(1, ch));
                } else {
                    thisCommand.push_back(string_t());
                }
                hadLE = true;
                s = chClass;
            }
            break;

         case DQVar:
         case WordVar:
            if (ch == '{') {
                varName.clear();
                s = (s == DQVar ? DQVarBrace : WordVarBrace);
            } else if (chClass == Word) {
                thisCommand.back().append(environment[string_t(1, ch)]);
                s = (s == DQVar ? DQ : Word);
            } else {
                throw std::runtime_error("Invalid variable reference");
            }
            break;

         case DQVarBrace:
         case WordVarBrace:
            if (ch == '}') {
                thisCommand.back().append(environment[varName]);
                s = (s == DQVarBrace ? DQ : Word);
            } else {
                varName += char(ch);
            }
            break;

         case Brace:
            if (ch == '}' && braceLevel == 0) {
                s = Word;
            } else {
                if (ch == '"') {
                    s = BraceDQ;
                } else if (ch == '\'') {
                    s = BraceSQ;
                } else if (ch == '}') {
                    --braceLevel;
                } else if (ch == '{') {
                    ++braceLevel;
                }
                thisCommand.back() += char(ch);
            }
            break;

         case BraceDQ:
            thisCommand.back() += char(ch);
            if (ch == '\\') {
                s = BraceDQQ;
            } else if (ch == '"') {
                s = Brace;
            }
            break;

         case BraceDQQ:
            thisCommand.back() += char(ch);
            s = BraceDQ;
            break;

         case BraceSQ:
            thisCommand.back() += char(ch);
            if (ch == '\'') {
                s = Brace;
            }
            break;

         case Pipe:
            break;
        }

        if (s == Pipe) {
            // Prepare
            handlePipe(thisCommand, previousPipeResult);

            // Execute
            std::auto_ptr<IntValue> val(evalCommand(thisCommand));
            thisCommand.clear();

            // Postprocess
            if (IntArray* a = dynamic_cast<IntArray*>(val.get())) {
                toStringListResult(a, previousPipeResult);
            } else {
                previousPipeResult.push_back(toStringResult(val.get()));
            }
            s = Blank;
        }
    }

    // End of line
    switch (s) {
     case Blank:
        break;

     case Word:
        handleLE(thisCommand, hadLE);
        break;

     case DQ:
     case DQQ:
     case DQX:
     case DQU:
     case SQ:
     case BraceDQ:
     case BraceDQQ:
     case BraceSQ:
        throw std::runtime_error("Quote not closed at end of line");

     case WordVar:
     case WordVarBrace:
     case DQVar:
     case DQVarBrace:
        throw std::runtime_error("Variable reference not finished at end of line");

     case LE:
        throw std::runtime_error("Expecting name after '<'");

     case Pipe:
        throw std::runtime_error("Pipe not allowed at end of line");

     case Brace:
        if (terminal_mode >= ColorMode) {
            std::cout << "\033[36m(continue...)>\033[0m ";
        } else if (terminal_mode >= DumbMode) {
            std::cout << "(continue...)> ";
        }
        if (!getline(in, line)) {
            throw std::runtime_error("Unexpected end of file");
        }
        if (!thisCommand.back().empty()) {
            thisCommand.back() += "\n";
        }
        goto again;
    }

    // Quick exit
    if (!hadNonblank) {
        return false;
    }

    // Prepare
    handlePipe(thisCommand, previousPipeResult);

    // Execute
    lastResult.reset(evalCommand(thisCommand));
    return true;
}

static bool
evalStream(std::istream& in, TerminalMode terminal_mode)
{
    std::auto_ptr<IntValue> lastResult;
    while (!contexts.empty()) {
        if (terminal_mode >= DumbMode) {
            if (terminal_mode >= ColorMode) {
                std::cout << "\033[36;1m";
            }
            for (uint32_t i = 0; i < contexts.size(); ++i) {
                if (i != 0) {
                    std::cout << " ";
                }
                std::cout << contexts[i]->getPrefix();
            }
            if (terminal_mode >= ColorMode) {
                std::cout << ">\033[0m ";
            } else {
                std::cout << "> ";
            }
        }
        string_t line;
        if (!getline(in, line)) {
            break;
        }
        try {
            if (evalLine(in, terminal_mode, line, lastResult) && lastResult.get()) {
                std::cout << "result=";
                if (terminal_mode >= ColorMode) {
                    std::cout << "\033[32;1m";
                }
                showValue(lastResult.get(), string_t());
                if (terminal_mode >= ColorMode) {
                    std::cout << "\033[0m";
                }
                std::cout << "\n";
            }
        }
        catch (std::exception& e) {
            if (terminal_mode >= ColorMode) {
                std::cout << "ERROR: \033[31;1m" << e.what() << "\033[0m\n";
            } else if (terminal_mode >= DumbMode) {
                std::cout << "ERROR: " << e.what() << "\n";
            } else {
                std::cout << "ERROR: " << e.what() << "\n";
                std::cout << "ERROR:   in command '" << line << "'\n";
            }
        }
    }
    return getBoolValue(lastResult.get()) > 0;
}

IntValue*
evalSequence(string_t s)
{
    // Must attach a "\n" so we can safely use getline.
    std::stringstream ss(toStandardString(s) + "\n");
    ss.seekg(0);

    // Do it
    std::auto_ptr<IntValue> lastResult;
    string_t line;
    while (!contexts.empty() && getline(ss, line)) {
        try {
            evalLine(ss, PipeMode, line, lastResult);
        }
        catch (...) {
            std::cout << "ERROR: in nested command '" << line << "':\n";
            throw;
        }
    }
    return lastResult.release();
}

bool
evalSequenceToBool(string_t s)
{
    std::auto_ptr<IntValue> lastResult(evalSequence(s));
    return getBoolValue(lastResult.get()) > 0;
}


int main(int /*argc*/, char** argv)
{
    loadConfiguration(checkConfig, true);
    contexts.push_back(&root);

    StringList_t list;
    while (const char* p = *++argv) {
        if (*p == '-') {
            // "--foo" and "-foo" are the same
            if (p[1] == '-') {
                ++p;
            }

            if (strcmp(p, "-help") == 0) {
                std::cout << "c2console - PlanetsCentral Console\n\n"
                          << "c2console [VAR=VALUE] [--OPTION=VALUE] [COMMAND ARGS...]\n\n"
                          << "VAR=VALUE        sets environment variables accessible as '${VAR}' in commands.\n"
                          << "--OPTION=VALUE   overrides config file settings.\n"
                          << "COMMAND ARGS...  is executed as if entered at the regular prompt\n\n"
                          << "If no COMMAND is given, c2console is interactive.\n";
                return 0;
            } else {
                const char* q = strchr(p, '=');
                if (q != 0 && checkConfig1(strUCase(string_t(p+1, q-p-1)), q+1)) {
                    // ok
                } else {
                    std::cerr << "c2console: unknown option '" << p << "'\n";
                }
            }
        } else if (const char* q = strchr(p, '=')) {
            environment[string_t(p, q-p)] = q+1;
        } else {
            break;
        }
    }

    if (*argv != 0) {
        list.push_back(*argv);
        while (const char* p = *++argv) {
            list.push_back(p);
        }
        try {
            std::auto_ptr<IntValue> val(evalCommand(list));
            if (val.get()) {
                std::cout << "result=";
                showValue(val.get(), string_t());
                std::cout << "\n";
            }
        }
        catch (std::exception& e) {
            std::cout << "ERROR: " << e.what() << "\n";
            return 1;
        }
    } else {
        if (isatty(0)) {
            char* t = getenv("TERM");
            if (t == 0 || strstr(t, "emacs") != 0) {
                terminal_mode = DumbMode;
                std::cout << "|\n"
                    "|  PlanetsCentral Console\n"
                    "|\n";
            } else {
                terminal_mode = ColorMode;
                std::cout << "\033[33;1m|\n"
                    "|  PlanetsCentral Console\n"
                    "|\033[0m\n";
            }
        } else {
            terminal_mode = PipeMode;
        }
        evalStream(std::cin, terminal_mode);
    }
}
