/**
  *  \file planetscentral/net/url.cc
  *  \brief Networking: URL
  */

#include "planetscentral/net/url.h"
#include "planetscentral/net/encoding.h"

/** Constructor. */
Url::Url()
    : m_Port(0), m_Valid(false)
{ }

/** Destructor. */
Url::~Url()
{ }

/** Parse an URL. If the URL is syntactically valid, store it in this
    object. Otherwise, does not modify it.
    \param s URL string
    \return true on success */
bool
Url::parse(const string_t s)
{
    string_t::size_type n = s.find(':');
    if (n == s.npos)
        /* No protocol identifier */
        return false;
    if (s.substr(n, 3) != "://")
        /* No "://" */
        return false;

    /* ok, found a protocol */
    string_t proto(s, 0, n);
    n += 3;

    string_t::size_type userend = s.find_first_of("@:/?#", n);
    string_t user;
    if (userend != s.npos && s[userend] == '@') {
        user.assign(s, n, userend-n);
        n = userend+1;
    }

    string_t::size_type hostend = s.find_first_of(":/?#", n);
    string_t::size_type portend = hostend;
    int port = -1;
    if (hostend != s.npos && s[hostend] == ':') {
        ++portend;
        port = 0;
        while (portend < s.size() && s[portend] >= '0' && s[portend] <= '9') {
            port = 10*port + s[portend] - '0';
            portend++;
        }
        if (portend == hostend+1)
            port = -1;
    }

    /* now, portend must point to the end of the string or at the '/'
       that begins the file name */
    if (portend < s.size()) {
        if (s[portend] == '?' || s[portend] == '#') {
            /* http://localhost?qry */
            m_Path.assign("/");
            m_Path.append(s, portend, s.npos);
        } else if (s[portend] == '/') {
            m_Path.assign(s, portend, s.npos);
        } else {
            return false;
        }
    } else {
        m_Path.assign("/");
    }
    m_Protocol = proto;
    m_Host.assign(s, n, hostend-n);
    m_Port = port;
    m_User = user;
    m_Valid = true;
    return true;
}

/** Convert content of this object back into string. Usually returns
    the same string as given to parse(). */
string_t
Url::toString() const
{
    if (!m_Valid)
        return string_t();

    string_t port;
    if (m_Port >= 0)
        port = ":" + itoa(m_Port);

    string_t user;
    if (!m_User.empty()) {
        user = m_User + '@';
    }

    return m_Protocol + "://" + user + m_Host + port + m_Path;
}

/** Get base name of URL. For http://host/dir/file, returns file.
    Returns "index.html" if URL contains no file part. */
string_t
Url::getBaseName() const
{
    if (!m_Valid)
        return string_t();

    string_t path = m_Path;
    string_t::size_type n = path.find_first_of("?#");
    if (n != path.npos)
        path.erase(n);

    n = path.rfind('/');
    if (n == string_t::npos || n+1 == path.size())
        return string_t("index.html");
    else
        return string_t(path, n+1);
}

// /** Remove parameters from this URL. If this URL contains parameters
//     (query), remove those and add them to \c e. */
// void
// Url::trimParameters(KVEnumerator& e)
// {
//     trimParameters(m_Path, e);
// }

// /** Remove parameters from string. If this string contains a '?',
//     removes the query string after it and assigns the query content to
//     \c e. */
// void
// Url::trimParameters(string_t& s, KVEnumerator& e)
// {
//     string_t::size_type beg = s.find('?');
//     if (beg == s.npos)
//         return;

//     /* we have parameters */
//     string_t::size_type n = beg;
//     while (n != s.npos) {
//         string_t::size_type end = s.find('&', n+1);
//         string_t here(s, n+1, end == s.npos ? s.npos : end - n - 1);
//         string_t::size_type eq = here.find('=');
//         if (eq != s.npos) {
//             e.handleItem(decodeUrl(here.substr(0, eq)),
//                          string_t(),
//                          decodeUrl(here.substr(eq+1)));
//         }
//         n = end;
//     }
//     s.erase(beg);
// }
