/**
  *  \file planetscentral/talk/bbout.cc
  *  \brief Talk: BBCode output
  */

#include "planetscentral/talk/bbout.h"
#include "planetscentral/talk/text.h"

static void renderPG(TextNode* n, string_t& result);
static void renderInline(TextNode* n, string_t& result);

static const char*
getInlineTagName(uint8_t minor)
{
    switch (minor) {
     case miInBold:             return "b";
     case miInItalic:           return "i";
     case miInStrikeThrough:    return "s";
     case miInUnderline:        return "u";
     case miInMonospace:        return "tt";
     default:                   return 0;
    }
}

static const char*
getInlineAttrTagName(uint8_t minor)
{
    switch (minor) {
     case miIAColor:   return "color";
     case miIASize:    return "size";
     case miIAFont:    return "font";
     default:          return 0;
    }
}

static const char*
getLinkTagName(uint8_t minor)
{
    switch (minor) {
     case miLinkUrl:       return "url";
     case miLinkEmail:     return "email";
     case miLinkThread:    return "thread";
     case miLinkPost:      return "post";
     case miLinkGame:      return "game";
     case miLinkUser:      return "user";
     case miLinkForum:     return "forum";
     case miLinkImage:     return "img";
     default:              return 0;
    }
}

static void
renderChildrenPG(TextNode* n, string_t& result)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        if (i != 0) {
            result += "\n\n";
        }
        renderPG(n->children[i], result);
    }
}

static void
renderChildrenInline(TextNode* n, string_t& result)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        renderInline(n->children[i], result);
    }
}

static void
renderAttr(TextNode* n, string_t& result)
{
    if (!n->text.empty()) {
        result += "=";
        if (n->text.find_first_of("[]") != string_t::npos) {
            result += "\"";
            result += n->text;
            result += "\"";
        } else {
            result += n->text;
        }
    }
}

static void
renderInline(TextNode* n, string_t& result)
{
    switch (n->major) {
     case maPlain:
        // FIXME: quote
        result += n->text;
        break;
     case maInline:
        if (const char* tag = getInlineTagName(n->minor)) {
            result += "[";
            result += tag;
            result += "]";
            renderChildrenInline(n, result);
            result += "[/";
            result += tag;
            result += "]";
        } else {
            renderChildrenInline(n, result);
        }
        break;
     case maInlineAttr:
        if (const char* tag = getInlineAttrTagName(n->minor)) {
            result += "[";
            result += tag;
            renderAttr(n, result);
            result += "]";
            renderChildrenInline(n, result);
            result += "[/";
            result += tag;
            result += "]";
        } else {
            renderChildrenInline(n, result);
        }
        break;
     case maLink:
        if (const char* tag = getLinkTagName(n->minor)) {
            if (n->children.empty()
                && n->text.find_first_of("[]") == string_t::npos
                && (n->text.empty() || n->text[0] != '@')
                && (n->text.find('@') == string_t::npos || n->text.find_first_of(" \t\n\r") == string_t::npos))
            {
                /* The above condition is true if we have no children, and we are reasonably
                   sure that the content of the attribute needs no escaping, so we can use
                   the abbreviated format:
                   - no []
                   - does not start with @
                   - does not contain both @ and whitespace (this would mean we risk
                     generating an at-link) */
                result += "[";
                result += tag;
                result += "]";
                result += n->text;
                result += "[/";
                result += tag;
                result += "]";
            } else {
                result += "[";
                result += tag;
                renderAttr(n, result);
                result += "]";
                renderChildrenInline(n, result);
                result += "[/";
                result += tag;
                result += "]";
            }
        } else {
            renderChildrenInline(n, result);
        }
        break;

     default:
        break;
    }
}

static void
renderPG(TextNode* n, string_t& result)
{
    switch (n->major) {
     case maGroup:
        switch (n->minor) {
         case miGroupQuote:
            result += "[quote";
            renderAttr(n, result);
            result += "]\n";
            renderChildrenPG(n, result);
            result += "[/quote]";
            break;
         case miGroupList:
            result += "[list";
            renderAttr(n, result);
            result += "]\n";
            renderChildrenPG(n, result);
            result += "[/list]";
            break;
         case miGroupListItem:
            result += "[*] ";
            renderChildrenPG(n, result);
            break;
         case miGroupRoot:
         default:
            renderChildrenPG(n, result);
            break;
        }
        break;
        
     case maParagraph:
        switch (n->minor) {
         case miParCode:
            result += "[code";
            renderAttr(n, result);
            result += "]\n";
            renderChildrenInline(n, result);
            result += "[/code]";
            break;
         case miParCentered:
            result += "[center";
            renderAttr(n, result);
            result += "]\n";
            renderChildrenInline(n, result);
            result += "[/center]";
            break;
         default:
         case miParNormal:
            renderChildrenInline(n, result);
            break;
        }
     default:
        break;
    }
}

string_t
renderBB(TextNode* n)
{
    string_t result;
    renderChildrenPG(n, result);
    return result;
}
